Index: pbx.c
===================================================================
RCS file: /usr/cvsroot/asterisk/pbx.c,v
retrieving revision 1.173
diff -u -r1.173 pbx.c
--- pbx.c	19 Nov 2004 05:18:10 -0000	1.173
+++ pbx.c	22 Nov 2004 21:38:02 -0000
@@ -4979,20 +4979,26 @@
 			mode = 0;
 	} 
 	/* At this point we have a priority and maybe an extension and a context */
-	if (mode)
-		chan->priority += mode * ipri - 1;
-	else
-		chan->priority = ipri - 1;
-	if (exten && strcasecmp(exten, "BYEXTENSION"))
-		strncpy(chan->exten, exten, sizeof(chan->exten)-1);
-	if (context)
-		strncpy(chan->context, context, sizeof(chan->context)-1);
+
+	if (exten && !strcasecmp(exten, "BYEXTENSION"))
+		exten = NULL;
+
+	if (mode) {
+		ipri = chan->priority + (ipri * mode);
+	}
+
+	if (ast_goto_if_exists(chan, context, exten, ipri)) {
+		ast_log(LOG_WARNING, "Error: nonexistant extension %s encountered in context %s\n", exten, context);
+		return -1;
+	}
+
 	if (option_verbose > 2)
 		ast_verbose( VERBOSE_PREFIX_3 "Goto (%s,%s,%d)\n", chan->context,chan->exten, chan->priority+1);
 	ast_cdr_update(chan);
 	return 0;
 }
 
+
 int pbx_builtin_serialize_variables(struct ast_channel *chan, char *buf, size_t size) 
 {
 	struct ast_var_t *variables;
@@ -5468,4 +5474,30 @@
 		}
 	return res;
 }
+
+
+int ast_goto_if_exists(struct ast_channel *chan, char* context, char *exten, int priority) 
+{
+	if(chan) {
+		if(ast_exists_extension(chan, 
+								context ? context : chan->context,
+								exten ? exten : chan->exten,
+								priority ? priority : chan->priority,
+								chan->cid.cid_num)) {
+			return ast_async_goto(chan,
+								  context ? context : chan->context,
+								  exten ? exten : chan->exten,
+								  priority ? priority : chan->priority);
+		} else 
+			return -3;
+	}
+	
+	return -2;
+}
+
+int ast_parsable_goto(struct ast_channel *chan, char *goto_string) 
+{
+	return pbx_builtin_goto(chan, goto_string);
+}
+
 
Index: include/asterisk/pbx.h
===================================================================
RCS file: /usr/cvsroot/asterisk/include/asterisk/pbx.h,v
retrieving revision 1.33
diff -u -r1.33 pbx.h
--- include/asterisk/pbx.h	13 Nov 2004 22:44:33 -0000	1.33
+++ include/asterisk/pbx.h	22 Nov 2004 21:38:03 -0000
@@ -578,7 +578,8 @@
   set to 1, sets to auto fall through.  If newval set to 0, sets to no auto
   fall through (reads extension instead).  Returns previous value. */
 extern int pbx_set_autofallthrough(int newval);
-
+int ast_goto_if_exists(struct ast_channel *chan, char* context, char *exten, int priority);
+int ast_parsable_goto(struct ast_channel *chan, char *goto_string);
 #if defined(__cplusplus) || defined(c_plusplus)
 }
 #endif
Index: apps/app_dial.c
===================================================================
RCS file: /usr/cvsroot/asterisk/apps/app_dial.c,v
retrieving revision 1.107
diff -u -r1.107 app_dial.c
--- apps/app_dial.c	21 Nov 2004 20:38:32 -0000	1.107
+++ apps/app_dial.c	22 Nov 2004 21:38:04 -0000
@@ -68,7 +68,13 @@
 "             that are assigned to you.\n"
 "      'r' -- indicate ringing to the calling party, pass no audio until answered.\n"
 "      'm' -- provide hold music to the calling party until answered.\n"
-"      'M(x) -- Executes the macro (x) upon connect of the call\n"
+"      'M(x[^arg]) -- Executes the macro (x with ^ delim arg list) upon connect of the call.\n"
+"                     Also, the macro can set the MACRO_RESULT variable to do the following:\n"
+"                     -- ABORT - Hangup both legs of the call.\n"
+"                     -- CONGESTION - Behave as if line congestion was encountered.\n"
+"                     -- BUSY - Behave as if a busy signal was encountered. (n+101)\n"
+"                     -- CONTINUE - Hangup the called party and continue on in the dialplan.\n"
+"                     -- GOTO:<context>^<exten>^<priority> - Transfer the call.\n"
 "      'h' -- allow callee to hang up by hitting *.\n"
 "      'H' -- allow caller to hang up by hitting *.\n"
 "      'C' -- reset call detail record for this call.\n"
@@ -484,7 +490,7 @@
 	char *sdtmfptr;
 	char sdtmfdata[256] = "";
 	char *stack,*var;
-	char *mac = NULL, macroname[256] = "";
+	char *mac = NULL, *macroname = NULL;
 	char status[256]="";
 	char toast[80];
 	int play_to_caller=0,play_to_callee=0;
@@ -492,9 +498,12 @@
 	char *varname;
 	int vartype;
 	char *outbound_group = NULL;
-
+	char *macro_result = NULL, *macro_transfer_exten = NULL, *macro_transfer_context = NULL;
 	int digit = 0;
 	time_t start_time, answer_time, end_time;
+	struct ast_app *app = NULL;
+	int mpri = 0;
+	char *mpri_ptr = NULL;
 
 	if (!data) {
 		ast_log(LOG_WARNING, "Dial requires an argument (technology1/number1&technology2/number2...|optional timeout|options)\n");
@@ -661,7 +670,7 @@
 		/* Get the macroname from the dial option string */
 		if ((mac = strstr(transfer, "M("))) {
 			hasmacro = 1;
-			strncpy(macroname, mac + 2, sizeof(macroname) - 1);
+			macroname = ast_strdupa(mac + 2);
 			while (*mac && (*mac != ')'))
 				*(mac++) = 'X';
 			if (*mac)
@@ -1001,8 +1010,6 @@
 			res = 0;
 
 		if (hasmacro && macroname) {
-			void *app = NULL;
-
 			res = ast_autoservice_start(chan);
 			if (res) {
 				ast_log(LOG_ERROR, "Unable to start autoservice on calling channel\n");
@@ -1012,6 +1019,9 @@
 			app = pbx_findapp("Macro");
 
 			if (app && !res) {
+				for(res=0;res<strlen(macroname);res++)
+					if(macroname[res] == '^')
+						macroname[res] = '|';
 				res = pbx_exec(peer, app, macroname, 1);
 				ast_log(LOG_DEBUG, "Macro exited with status %d\n", res);
 				res = 0;
@@ -1023,6 +1033,70 @@
 			if (ast_autoservice_stop(chan) < 0) {
 				ast_log(LOG_ERROR, "Could not stop autoservice on calling channel\n");
 				res = -1;
+			}
+
+			if (!res) {
+				if ((macro_result = pbx_builtin_getvar_helper(peer, "MACRO_RESULT"))) {
+					if (!strcasecmp(macro_result, "BUSY")) {
+						strncpy(status, macro_result, sizeof(status) - 1);
+						if (!ast_goto_if_exists(chan, NULL, NULL, chan->priority + 101)) {
+							go_on = 1;
+						}
+						res = -1;
+					}
+					else if (!strcasecmp(macro_result, "CONGESTION") || !strcasecmp(macro_result, "CHANUNAVAIL")) {
+						strncpy(status, macro_result, sizeof(status) - 1);
+						go_on = 1;
+						res = -1;
+					}
+					else if (!strcasecmp(macro_result, "CONTINUE")) {
+						/* hangup peer and keep chan alive assuming the macro has changed 
+						   the context / exten / priority or perhaps 
+						   the next priority in the current exten is desired.
+						*/
+						go_on = 1;
+						res = -1;
+					} else if (!strcasecmp(macro_result, "ABORT")) {
+						/* Hangup both ends unless the caller has the g flag */
+						res = -1;
+					} else if(!strncasecmp(macro_result, "GOTO:",5) && (macro_transfer_exten = ast_strdupa(macro_result + 5))) {
+						res = -1;
+						/* perform a transfer to a new extension */
+						if(strchr(macro_transfer_exten,'^')) { /* context^exten^priority*/
+							/* no brainer mode... substitute ^ with | and feed it to builtin goto */
+							for(res=0;res<strlen(macro_transfer_exten);res++)
+								if(macro_transfer_exten[res] == '^')
+									macro_transfer_exten[res] = '|';
+
+							if(!ast_parsable_goto(chan, macro_transfer_exten))
+								go_on = 1;
+							
+						} else {
+							/* Support anthm's proposed general syntax */
+							/* 1234 or 1234:1 or 1234@mycon or 1234@mycon:1 */
+							mpri = 1;
+							if ((macro_transfer_context = strchr(macro_transfer_exten,'@'))) {
+								*macro_transfer_context = '\0';
+								macro_transfer_context++;
+								if((mpri_ptr = strchr(macro_transfer_context,':'))) {
+									*mpri_ptr = '\0';
+									mpri_ptr++;
+									mpri = atoi(mpri_ptr);
+								}
+							} else {
+								macro_transfer_context = chan->context;
+								if((mpri_ptr = strchr(macro_transfer_exten,':'))) {
+									*mpri_ptr = '\0';
+									mpri_ptr++;
+									mpri = atoi(mpri_ptr);
+								}
+							}
+						
+							if (!ast_goto_if_exists(chan, macro_transfer_context, macro_transfer_exten, mpri)) 
+								go_on = 1;
+						}
+					}
+				}
 			}
 		}
 
